//
//  ViewController.m
//  AudioTalk
//
//  Created by 孙震 on 2022/5/23.
//

#import "ViewController.h"
#import <AVFoundation/AVFoundation.h>
#import "G711/g711_table.h"
#import "AuthUtil.h"
#import "QRDMicrophoneSource.h"

/**
 api文档:https://developer.qiniu.com/qvs/8158/voice-call
 token鉴权文档：https://developer.qiniu.com/qvs/6713/authentication
 */

#error 这里替换参数
#define QiNiuToken   @"your qiniu token"
#define nameSpaceId  @"your namespace id"
#define deviceId     @"your device id"

#define bufferLength 1*1024*1024

@interface ViewController ()<QRDMicrophoneSourceDelegate>
@property (weak, nonatomic) IBOutlet UIButton *fetchUrlBtn;
@property (weak, nonatomic) IBOutlet UIButton *recorderBtn;
@property (weak, nonatomic) IBOutlet UILabel *statusLabel;
  
@property (strong ,nonatomic) NSString *talkUrl; //通话的url
@property (nonatomic) dispatch_queue_t readQueue;
//@property (strong, nonatomic) NSMutableData *pcmData;
@property (strong, nonatomic) QRDMicrophoneSource *microPhone;
@end

@implementation ViewController
{
    char buffer[bufferLength];
    int  pcmWriteOffset;
    int  pcmReadOffset;
}

- (void)viewDidLoad {
    [super viewDidLoad];
    [self setup];
    
}

- (void)setup {
    self.microPhone = [[QRDMicrophoneSource alloc] init];
    self.microPhone.delegate = self;
    self.readQueue = dispatch_queue_create("com.qiniu.qrd.pcmRead", DISPATCH_QUEUE_SERIAL);
    self.recorderBtn.enabled = NO;
    [self getTalkUrlBtnClicked:self.fetchUrlBtn];
  
}
 

#pragma mark -- btn click
- (IBAction)recordButtonClicked:(UIButton *)sender {
    if (self.microPhone.isRunning) {
        //停止录制
        [self.microPhone stopRunning];
        [self.recorderBtn setTitle:@"开始通话" forState:UIControlStateNormal];
         
    } else {
        //开始录制
        pcmReadOffset = 0;
        pcmWriteOffset = 0;
        [self.microPhone startRunning];
        [self.recorderBtn setTitle:@"结束通话" forState:UIControlStateNormal];
        dispatch_async(self.readQueue, ^{
            while (true) {
                [NSThread sleepForTimeInterval:0.5];
                if (!self.microPhone.isRunning) {
                    break;
                }
                NSData *data = [self readPcm];
                if (data && self.talkUrl) {
                    NSData *g711 = [self Pcm2G711a:data];
                    //发送
                    [self postData:g711 url:self.talkUrl complete:^(int code) {
                        if (code  == 0) {
                            NSLog(@"成功发送语音");
                        } else if(code == 6002) {
                            self.talkUrl = nil;
                            NSLog(@"error:talk地址已经失效，请重新获取!");
                            [self.microPhone stopRunning];
                        } else {
                            self.talkUrl = nil;
                            [self.microPhone stopRunning];
                            NSLog(@"error:其他错误");
                        }
                    }];
                } else {
                    if(self.talkUrl) {
                        NSLog(@"error:请先获取通话地址！！！");
                    }
                    if(data) {
                        NSLog(@"error:没有获取到音频数据！！！");
                    }
                }
            }
        });
    }
}
 
#pragma mark - PCM2G711A
// pcm转成g711a
// pcm 的采样率 8k  16bit
// g711a 采样率 8k  8bit
- (NSData *)Pcm2G711a:(NSData *)pcmData {
    NSLog(@"convert g711  start");
    char *pcmCdata = (char *)[pcmData bytes];
    char *g711data = malloc(pcmData.length/2);
    pcm16_alaw_tableinit();
    pcm16_to_alaw((int)pcmData.length, pcmCdata, g711data);
    
    NSData *data = [NSData dataWithBytes:g711data length:pcmData.length/2];
    NSString *path = [NSTemporaryDirectory() stringByAppendingString:@"recording.g711"];
    
    if([[NSFileManager defaultManager] fileExistsAtPath:path]) {
        [[NSFileManager defaultManager] removeItemAtPath:path error:nil];
    }
    [data writeToFile:path atomically:YES];
    
    NSLog(@"convert g711  end");
    
    return data;
}

#pragma mark - sendData

- (IBAction)getTalkUrlBtnClicked:(id)sender {
    [self getTalkUrl:^(NSString *url) {
        dispatch_async(dispatch_get_main_queue(), ^{
            if (url) {
                self.talkUrl = url;
                [self setStatus:@"获取通话连接成功"];
                self.recorderBtn.enabled = YES;
            } else {
                [self setStatus:@"获取通话连接失败"];
                self.recorderBtn.enabled = NO;
            }
        });
    }];
    
}

- (void)getTalkUrl:(void(^)(NSString *url))result {
    /**
     api文档:https://developer.qiniu.com/qvs/8158/voice-call
     token鉴权文档：https://developer.qiniu.com/qvs/6713/authentication
     */
    //如果当前有值 则不需要请求 1分钟内不发送数据才失效
    if(self.talkUrl){
        result(self.talkUrl);
        return;
    }
     
    NSString *urlStr = [NSString stringWithFormat:@"http://qvs.qiniuapi.com/v1/namespaces/%@/devices/%@/talk",nameSpaceId,deviceId];
    NSMutableURLRequest *request = [[NSMutableURLRequest alloc] initWithURL:[NSURL URLWithString:urlStr]];
    [request setHTTPMethod:@"POST"];
    [request setValue:@"application/json" forHTTPHeaderField:@"Content-Type"];
    [request setValue:QiNiuToken forHTTPHeaderField:@"Authorization"];
    NSData *body = [@"{\"isV2\":true}" dataUsingEncoding:NSUTF8StringEncoding];
    [request setHTTPBody:body];
    NSURLSessionDataTask *task = [[NSURLSession sharedSession] dataTaskWithRequest:request completionHandler:^(NSData * _Nullable data, NSURLResponse * _Nullable response, NSError * _Nullable error) {
                NSDictionary *resp = [NSJSONSerialization JSONObjectWithData:data options:NSJSONReadingFragmentsAllowed  error:nil];
        NSLog(@"%@",resp);
        NSString *httpUrl = resp[@"audioSendAddrForHttp"];
        result(httpUrl);
    }];
    
    [task resume];
}

//发送数据
-(void)postData:(NSData *)data url:(NSString *)url complete:(void(^)(int code))result {
    self.talkUrl = url;
    NSMutableURLRequest *request = [[NSMutableURLRequest alloc] initWithURL:[NSURL URLWithString:url]];
    [request setHTTPMethod:@"POST"];
    [request setValue:@"application/json" forHTTPHeaderField:@"content-type"];
    //data to base64
    NSString *dataStr = [AuthUtil data2SBase64:data];
    NSLog(@"datastr:%lu",(unsigned long)dataStr.length);
    NSDictionary *param = [NSDictionary dictionaryWithObject:dataStr forKey:@"base64_pcm"];
    NSData *json = [NSJSONSerialization dataWithJSONObject:param options:NSJSONWritingPrettyPrinted error:nil];
    [request setHTTPBody:json];
    NSURLSessionDataTask *task = [[NSURLSession sharedSession] dataTaskWithRequest:request completionHandler:^(NSData * _Nullable data, NSURLResponse * _Nullable response, NSError * _Nullable error) {
        if(error) {
            NSLog(@"请求报错:%@",error);
            return;
        }
        if(data) {
            NSDictionary *resp = [NSJSONSerialization JSONObjectWithData:data options:NSJSONReadingFragmentsAllowed  error:nil];
            result([resp[@"code"] intValue]);
        }
    }];
    
    [task resume];
}

- (void)setStatus:(NSString *)string {
    self.statusLabel.text = string;
}

#pragma -mark 获取PCM

- (void)microphoneSource:(QRDMicrophoneSource *)source didGetAudioBuffer:(AudioBuffer *)audioBuffer {
    
    int length = audioBuffer->mDataByteSize;
    
    if(length+pcmWriteOffset <= bufferLength){
        memcpy(buffer + pcmWriteOffset, audioBuffer->mData, audioBuffer->mDataByteSize);
        pcmWriteOffset += length;
    } else {
        //buffer不够存放pcm数据 需要从头部开始
        //末尾剩余长度
        size_t endSize = bufferLength - pcmWriteOffset;
        //超出的长度
        size_t outSize = length + pcmWriteOffset - bufferLength;
        memcpy(buffer + pcmWriteOffset, audioBuffer->mData, endSize);
        pcmWriteOffset = 0;
        memcpy(buffer + pcmWriteOffset, (char *)(audioBuffer->mData) + endSize, outSize);
        pcmWriteOffset += outSize;
    }
}

- (NSData *)readPcm {
    if (pcmReadOffset == pcmWriteOffset == 0 && !self.microPhone.isRunning) {
        return nil;
    }
    
    if (pcmWriteOffset > pcmReadOffset) {
        NSUInteger size = pcmWriteOffset - pcmReadOffset;
        NSData * data = [NSData dataWithBytes:buffer + pcmReadOffset length:size];
        pcmReadOffset += size;
        return data;
    } else {
        NSUInteger size1 = bufferLength - pcmReadOffset;
        NSData *data1 = [NSData dataWithBytes:buffer + pcmReadOffset length:size1];
        pcmReadOffset = 0;
        NSUInteger size2  = pcmWriteOffset;
        NSData *data2 = [NSData dataWithBytes:buffer length:size2];
        pcmReadOffset += size2;
        NSMutableData *muteData = [[NSMutableData alloc] init];
        [muteData appendData:data1];
        [muteData appendData:data2];
        return muteData;
    }
}

@end
